import py
import pytest
import sys

pytest_plugins = 'doctest', 'pytester'

collect_ignore = ['build', 'doc/_build']


def pytest_addoption(parser):
    group = parser.getgroup("pylib", "py lib testing options")
    group.addoption('--runslowtests',
           action="store_true", dest="runslowtests", default=False,
           help=("run slow tests"))

@pytest.fixture
def sshhost(request):
    val = request.config.getvalue("sshhost")
    if val:
        return val
    py.test.skip("need --sshhost option")


# XXX copied from execnet's conftest.py - needs to be merged
winpymap = {
    'python2.7': r'C:\Python27\python.exe',
}


def getexecutable(name, cache={}):
    try:
        return cache[name]
    except KeyError:
        executable = py.path.local.sysfind(name)
        if executable:
            if name == "jython":
                import subprocess
                popen = subprocess.Popen(
                    [str(executable), "--version"],
                    universal_newlines=True, stderr=subprocess.PIPE)
                out, err = popen.communicate()
                if not err or "2.5" not in err:
                    executable = None
        cache[name] = executable
        return executable


@pytest.fixture(params=('python2.7', 'pypy-c', 'jython'))
def anypython(request):
    name = request.param
    executable = getexecutable(name)
    if executable is None:
        if sys.platform == "win32":
            executable = winpymap.get(name, None)
            if executable:
                executable = py.path.local(executable)
                if executable.check():
                    return executable
        py.test.skip("no %s found" % (name,))
    return executable



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from pathlib import Path
from unittest import mock
from unittest.mock import Mock, patch

import pytest
from mozbuild.base import MachCommandBase
from mozunit import main

import mach.decorators
import mach.registrar
from mach.base import MachError
from mach.decorators import Command, CommandArgument, SubCommand
from mach.requirements import MachEnvRequirements
from mach.site import CommandSiteManager, MozSiteMetadata, SitePackagesSource


@pytest.fixture
def registrar(monkeypatch):
    test_registrar = mach.registrar.MachRegistrar()
    test_registrar.register_category(
        "testing", "Mach unittest", "Testing for mach decorators"
    )
    monkeypatch.setattr(mach.decorators, "Registrar", test_registrar)
    return test_registrar


def test_register_command_with_argument(registrar):
    inner_function = Mock()
    context = Mock()
    context.cwd = "."

    @Command("cmd_foo", category="testing")
    @CommandArgument("--arg", default=None, help="Argument help.")
    def run_foo(command_context, arg):
        inner_function(arg)

    registrar.dispatch("cmd_foo", context, arg="argument")

    inner_function.assert_called_with("argument")


def test_register_command_with_metrics_path(registrar):
    context = Mock()
    context.cwd = "."

    metrics_path = "metrics/path"
    metrics_mock = Mock()
    context.telemetry.metrics.return_value = metrics_mock

    @Command("cmd_foo", category="testing", metrics_path=metrics_path)
    def run_foo(command_context):
        assert command_context.metrics == metrics_mock

    @SubCommand("cmd_foo", "sub_foo", metrics_path=metrics_path + "2")
    def run_subfoo(command_context):
        assert command_context.metrics == metrics_mock

    registrar.dispatch("cmd_foo", context)

    context.telemetry.metrics.assert_called_with(metrics_path)
    assert context.handler.metrics_path == metrics_path

    registrar.dispatch("cmd_foo", context, subcommand="sub_foo")
    assert context.handler.metrics_path == metrics_path + "2"


def test_register_command_sets_up_class_at_runtime(registrar):
    inner_function = Mock()

    context = Mock()
    context.cwd = "."

    # We test that the virtualenv is set up properly dynamically on
    # the instance that actually runs the command.
    @Command("cmd_foo", category="testing", virtualenv_name="env_foo")
    def run_foo(command_context):
        assert (
            Path(command_context.virtualenv_manager.virtualenv_root).name == "env_foo"
        )
        inner_function("foo")

    @Command("cmd_bar", category="testing", virtualenv_name="env_bar")
    def run_bar(command_context):
        assert (
            Path(command_context.virtualenv_manager.virtualenv_root).name == "env_bar"
        )
        inner_function("bar")

    def from_environment_patch(
        topsrcdir: str, state_dir: str, virtualenv_name, directory: str
    ):
        return CommandSiteManager(
            "",
            "",
            virtualenv_name,
            virtualenv_name,
            MozSiteMetadata(0, "mach", SitePackagesSource.VENV, "", ""),
            True,
            MachEnvRequirements(),
        )

    with mock.patch.object(
        CommandSiteManager, "from_environment", from_environment_patch
    ):
        with patch.object(MachCommandBase, "activate_virtualenv"):
            registrar.dispatch("cmd_foo", context)
            inner_function.assert_called_with("foo")
            registrar.dispatch("cmd_bar", context)
            inner_function.assert_called_with("bar")


def test_cannot_create_command_nonexisting_category(registrar):
    with pytest.raises(MachError):

        @Command("cmd_foo", category="bar")
        def run_foo(command_context):
            pass


def test_subcommand_requires_parent_to_exist(registrar):
    with pytest.raises(MachError):

        @SubCommand("sub_foo", "foo")
        def run_foo(command_context):
            pass


if __name__ == "__main__":
    main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import os
import platform
import re
import subprocess

import six

from .macintelpower import MacIntelPower
from .mozpowerutils import average_summary, frequency_summary, get_logger, sum_summary

OSCPU_COMBOS = {
    "darwin-intel": MacIntelPower,
}

SUMMARY_METHODS = {
    "utilization": average_summary,
    "power-usage": sum_summary,
    "power-watts": frequency_summary,
    "frequency-cpu": frequency_summary,
    "frequency-gpu": frequency_summary,
    "default": sum_summary,
}


class OsCpuComboMissingError(Exception):
    """OsCpuComboMissingError is raised when we cannot find
    a class responsible for the OS, and CPU combination that
    was detected.
    """

    pass


class MissingProcessorInfoError(Exception):
    """MissingProcessorInfoError is raised when we cannot find
    the processor information on the machine. This is raised when
    the file is missing (mentioned in the error message) or if
    an exception occurs when we try to gather the information
    from the file.
    """

    pass


class MozPower(object):
    """MozPower provides an OS and CPU independent interface
    for initializing, finalizing, and gathering power measurement
    data from OS+CPU combo-dependent measurement classes. The combo
    is detected automatically, and the correct class is instantiated
    based on the OSCPU_COMBOS list. If it cannot be found, an
    OsCpuComboMissingError will be raised.

    If a newly added power measurer does not have the required functions
    `initialize_power_measurements`, `finalize_power_measurements`,
    or `get_perfherder_data`, then a NotImplementedError will be
    raised.

    Android power measurements are currently not supported by this
    module.

    ::

       from mozpower import MozPower

       mp = MozPower(output_file_path='dir/power-testing')

       mp.initialize_power_measurements()
       # Run test...
       mp.finalize_power_measurements(test_name='raptor-test-name')

       perfherder_data = mp.get_perfherder_data()
    """

    def __init__(
        self,
        android=False,
        logger_name="mozpower",
        output_file_path="power-testing",
        **kwargs
    ):
        """Initializes the MozPower object, detects OS and CPU (if not android),
        and instatiates the appropriate combo-dependent class for measurements.

        :param bool android: run android power measurer.
        :param str logger_name: logging logger name. Defaults to 'mozpower'.
        :param str output_file_path: path to where output files will be stored.
            Can include a prefix for the output files, i.e. 'dir/raptor-test'
            would output data to 'dir' with the prefix 'raptor-test'.
            Defaults to 'power-testing', the current directory using
            the prefix 'power-testing'.
        :param dict kwargs: optional keyword arguments passed to power measurer.
        :raises: * OsCpuComboMissingError
                 * NotImplementedError
        """
        self.measurer = None
        self._os = None
        self._cpu = None
        self._logger = get_logger(logger_name)

        if android:
            self._logger.error("Android power usage measurer has not been implemented")
            raise NotImplementedError
        else:
            self._os = self._get_os().lower()
            cpu = six.text_type(self._get_processor_info().lower())

            if "intel" in cpu:
                self._cpu = "intel"
            else:
                self._cpu = "arm64"

        # OS+CPU combos are specified through strings such as 'darwin-intel'
        # for mac power measurement on intel-based machines. If none exist in
        # OSCPU_COMBOS, OsCpuComboMissingError will be raised.
        measurer = None
        oscpu_combo = "%s-%s" % (self._os, self._cpu)
        if oscpu_combo in OSCPU_COMBOS:
            measurer = OSCPU_COMBOS[oscpu_combo]
        else:
            raise OsCpuComboMissingError(
                "Cannot find OS+CPU combo for %s" % oscpu_combo
            )

        if measurer:
            self._logger.info(
                "Intializing measurer %s for %s power measurements, see below for errors..."
                % (measurer.__name__, oscpu_combo)
            )
            self.measurer = measurer(
                logger_name=logger_name, output_file_path=output_file_path, **kwargs
            )

    def _get_os(self):
        """Returns the operating system of the machine being tested. platform.system()
        returns 'darwin' on MacOS, 'windows' on Windows, and 'linux' on Linux systems.

        :returns: str
        """
        return platform.system()

    def _get_processor_info(self):
        """Returns the processor model type of the machine being tested.
        Each OS has it's own way of storing this information. Raises
        MissingProcessorInfoError if we cannot get the processor info
        from the expected locations.

        :returns: str
        :raises: * MissingProcessorInfoError
        """
        model = ""

        if self._get_os() == "Windows":
            model = platform.processor()

        elif self._get_os() == "Darwin":
            proc_info_path = "/usr/sbin/sysctl"
            command = [proc_info_path, "-n", "machdep.cpu.brand_string"]

            if not os.path.exists(proc_info_path):
                raise MissingProcessorInfoError(
                    "Missing processor info file for darwin platform, "
                    "expecting it here %s" % proc_info_path
                )

            try:
                model = subprocess.check_output(command).strip()
            except subprocess.CalledProcessError as e:
                error_log = str(e)
                if e.output:
                    error_log = e.output.decode()
                raise MissingProcessorInfoError(
                    "Error while attempting to get darwin processor information "
                    "from %s (exists) with the command %s: %s"
                    % (proc_info_path, str(command), error_log)
                )

        elif self._get_os() == "Linux":
            proc_info_path = "/proc/cpuinfo"
            model_re = re.compile(r""".*model name\s+[:]\s+(.*)\s+""")

            if not os.path.exists(proc_info_path):
                raise MissingProcessorInfoError(
                    "Missing processor info file for linux platform, "
                    "expecting it here %s" % proc_info_path
                )

            try:
                with open(proc_info_path) as cpuinfo:
                    for line in cpuinfo:
                        if not line.strip():
                            continue
                        match = model_re.match(line)
                        if match:
                            model = match.group(1)
                if not model:
                    raise Exception(
                        "No 'model name' entries found in the processor info file"
                    )
            except Exception as e:
                raise MissingProcessorInfoError(
                    "Error while attempting to get linux processor information "
                    "from %s (exists): %s" % (proc_info_path, str(e))
                )

        return model

    def initialize_power_measurements(self, **kwargs):
        """Starts the power measurements by calling the power measurer's
        `initialize_power_measurements` function.

        :param dict kwargs: keyword arguments for power measurer initialization
            function if they are needed.
        """
        if self.measurer is None:
            return
        self.measurer.initialize_power_measurements(**kwargs)

    def finalize_power_measurements(self, **kwargs):
        """Stops the power measurements by calling the power measurer's
        `finalize_power_measurements` function.

        :param dict kwargs: keyword arguments for power measurer finalization
            function if they are needed.
        """
        if self.measurer is None:
            return
        self.measurer.finalize_power_measurements(**kwargs)

    def get_perfherder_data(self):
        """Returns the partial perfherder data output produced by the measurer.
        For a complete perfherder data blob, see get_full_perfherder_data.

        :returns: dict
        """
        if self.measurer is None:
            return
        return self.measurer.get_perfherder_data()

    def _summarize_values(self, datatype, values):
        """Summarizes the given values based on the type of the
        data. See SUMMARY_METHODS for the methods used for each
        known data type. Defaults to using the sum of the values
        when a data type cannot be found.

        Data type entries in SUMMARY_METHODS are case-sensitive.

        :param str datastype: the measurement type being summarized.
        :param list values: the values to summarize.
        :returns: float
        """
        if datatype not in SUMMARY_METHODS:
            self._logger.warning(
                "Missing summary method for data type %s, defaulting to sum" % datatype
            )
            datatype = "default"

        summary_func = SUMMARY_METHODS[datatype]
        return summary_func(values)

    def get_full_perfherder_data(
        self, framework, lowerisbetter=True, alertthreshold=2.0
    ):
        """Returns a list of complete perfherder data blobs compiled from the
        partial perfherder data blob returned from the measurer. Each key entry
        (measurement type) in the partial perfherder data is parsed into its
        own suite within a single perfherder data blob.

        For example, a partial perfherder data blob such as:

        ::

           {
               'utilization': {<perfherder_data>},
               'power-usage': {<perfherder_data>}
           }

        would produce two suites within a single perfherder data blobs -
        one for utilization, and one for power-usage.

        Note that the 'values' entry must exist, otherwise the measurement
        type is skipped. Furthermore, if 'name', 'unit', or 'type' is missing
        we default to:

        ::

           {
               'name': 'mozpower',
               'unit': 'mWh',
               'type': 'power'
           }

        Subtests produced for each sub-suite (measurement type), have the naming
        pattern: <measurement_type>-<measured_name>

        Utilization of cpu would have the following name: 'utilization-cpu'
        Power-usage for cpu has the following name: 'power-usage-cpu'

        :param str framework: name of the framework being tested, i.e. 'raptor'.
        :param bool lowerisbetter: if set to true, low values are better than high ones.
        :param float alertthreshold: determines the crossing threshold at
            which an alert is generated.
        :returns: dict
        """
        if self.measurer is None:
            return

        # Get the partial data, and the measurers name for
        # logging purposes.
        partial_perfherder_data = self.get_perfherder_data()
        measurer_name = self.measurer.__class__.__name__

        suites = []
        perfherder_data = {"framework": {"name": framework}, "suites": suites}

        for measurement_type in partial_perfherder_data:
            self._logger.info("Summarizing %s data" % measurement_type)
            dataset = partial_perfherder_data[measurement_type]

            # Skip this measurement type if the 'values' entry
            # doesn't exist, and output a warning.
            if "values" not in dataset:
                self._logger.warning(
                    "Missing 'values' entry in partial perfherder data for measurement type %s "
                    "obtained from %s. This measurement type will not be processed."
                    % (measurement_type, measurer_name)
                )
                continue

            # Get the settings, if they exist, otherwise output
            # a warning and use a default entry.
            settings = {"test": "mozpower", "unit": "mWh", "type": "power"}

            for setting in settings:
                if setting in dataset:
                    settings[setting] = dataset[setting]
                else:
                    self._logger.warning(
                        "Missing '%s' entry in partial perfherder data for measurement type %s "
                        "obtained from %s, using %s as the default"
                        % (setting, measurement_type, measurer_name, settings[setting])
                    )

            subtests = []
            suite = {
                "name": "%s-%s" % (settings["test"], measurement_type),
                "type": settings["type"],
                "value": 0,
                "subtests": subtests,
                "lowerIsBetter": lowerisbetter,
                "unit": settings["unit"],
                "alertThreshold": alertthreshold,
            }

            # Parse the 'values' entries into subtests
            values = []
            for measure in dataset["values"]:
                value = dataset["values"][measure]
                subtest = {
                    "name": "%s-%s" % (measurement_type, measure),
                    "value": float(value),
                    "lowerIsBetter": lowerisbetter,
                    "alertThreshold": alertthreshold,
                    "unit": settings["unit"],
                }
                values.append((value, measure))
                subtests.append(subtest)

            # Summarize the data based on the measurement type
            if len(values) > 0:
                suite["value"] = self._summarize_values(measurement_type, values)
            suites.append(suite)

        return perfherder_data



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# add this directory to the path
import os
import sys

sys.path.append(os.path.dirname(__file__))

from session_store_test_case import SessionStoreTestCase


class TestSessionStoreEnabledAllWindows(SessionStoreTestCase):
    def setUp(self, include_private=True):
        """Setup for the test, enabling session restore.

        :param include_private: Whether to open private windows.
        """
        super(TestSessionStoreEnabledAllWindows, self).setUp(
            include_private=include_private, startup_page=3
        )

    def test_with_variety(self):
        """Test opening and restoring both standard and private windows.

        Opens a set of windows, both standard and private, with
        some number of tabs in them. Once the tabs have loaded, restarts
        the browser, and then ensures that the standard tabs have been
        restored, and that the private ones have not.
        """
        self.wait_for_windows(
            self.all_windows, "Not all requested windows have been opened"
        )

        self.marionette.quit()
        self.marionette.start_session()

        self.wait_for_windows(
            self.test_windows, "Non private browsing windows should have been restored"
        )

    def test_close_tabs(self):
        self.wait_for_windows(
            self.all_windows, "Not all requested windows have been opened"
        )

        self.close_all_tabs_and_restart()

        self.assertEqual(
            len(self.marionette.chrome_window_handles),
            1,
            msg="Windows from last session shouldn`t have been restored.",
        )
        self.assertEqual(
            len(self.marionette.window_handles),
            1,
            msg="Tabs from last session shouldn`t have been restored.",
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import sys
from pathlib import Path
from typing import Union
from unittest.mock import Mock

from mach.site import MozSiteMetadata, SitePackagesSource


class NoopTelemetry(object):
    def __init__(self, failed_glean_import):
        self._failed_glean_import = failed_glean_import

    def metrics(self, metrics_path: Union[str, Path]):
        return Mock()

    def submit(self, is_bootstrap):
        if self._failed_glean_import and not is_bootstrap:
            active_site = MozSiteMetadata.from_runtime()
            if active_site.mach_site_packages_source == SitePackagesSource.SYSTEM:
                hint = (
                    "Mach is looking for glean in the system packages. This can be "
                    "resolved by installing it there, or by allowing Mach to run "
                    "without using the system Python packages."
                )
            elif active_site.mach_site_packages_source == SitePackagesSource.NONE:
                hint = (
                    "This is because Mach is currently configured without a source "
                    "for native Python packages."
                )
            else:
                hint = "You may need to run |mach bootstrap|."

            print(
                f"Glean could not be found, so telemetry will not be reported. {hint}",
                file=sys.stderr,
            )


class GleanTelemetry(object):
    """Records and sends Telemetry using Glean.

    Metrics are defined in python/mozbuild/metrics.yaml.
    Pings are defined in python/mozbuild/pings.yaml.

    The "metrics" and "pings" properties may be replaced with no-op implementations if
    Glean isn't available. This allows consumers to report telemetry without having
    to guard against incompatible environments.

    Also tracks whether an employee was just automatically opted into telemetry
    during this mach invocation.
    """

    def __init__(
        self,
    ):
        self._metrics_cache = {}

    def metrics(self, metrics_path: Union[str, Path]):
        if metrics_path not in self._metrics_cache:
            from glean import load_metrics

            metrics = load_metrics(metrics_path)
            self._metrics_cache[metrics_path] = metrics

        return self._metrics_cache[metrics_path]

    def submit(self, _):
        from pathlib import Path

        from glean import load_pings

        pings = load_pings(Path(__file__).parent.parent / "pings.yaml")
        pings.usage.submit()



from __future__ import annotations

import sys


__all__ = ["asyncio_timeout"]


if sys.version_info[:2] >= (3, 11):
    from asyncio import timeout as asyncio_timeout  # noqa: F401
else:
    from .async_timeout import timeout as asyncio_timeout  # noqa: F401



# Copyright 2011, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""WebSocket opening handshake processor. This class try to apply available
opening handshake processors for each protocol version until a connection is
successfully established.
"""

from __future__ import absolute_import

import logging

from pywebsocket3 import common
from pywebsocket3.handshake import hybi
# Export AbortedByUserException, HandshakeException, and VersionException
# symbol from this module.
from pywebsocket3.handshake.base import (
    AbortedByUserException,
    HandshakeException,
    VersionException
)


_LOGGER = logging.getLogger(__name__)


def do_handshake(request, dispatcher):
    """Performs WebSocket handshake.

    Args:
        request: mod_python request.
        dispatcher: Dispatcher (dispatch.Dispatcher).

    Handshaker will add attributes such as ws_resource in performing
    handshake.
    """

    _LOGGER.debug('Client\'s opening handshake resource: %r', request.uri)
    # To print mimetools.Message as escaped one-line string, we converts
    # headers_in to dict object. Without conversion, if we use %r, it just
    # prints the type and address, and if we use %s, it prints the original
    # header string as multiple lines.
    #
    # Both mimetools.Message and MpTable_Type of mod_python can be
    # converted to dict.
    #
    # mimetools.Message.__str__ returns the original header string.
    # dict(mimetools.Message object) returns the map from header names to
    # header values. While MpTable_Type doesn't have such __str__ but just
    # __repr__ which formats itself as well as dictionary object.
    _LOGGER.debug('Client\'s opening handshake headers: %r',
                  dict(request.headers_in))

    handshakers = []
    handshakers.append(('RFC 6455', hybi.Handshaker(request, dispatcher)))

    for name, handshaker in handshakers:
        _LOGGER.debug('Trying protocol version %s', name)
        try:
            handshaker.do_handshake()
            _LOGGER.info('Established (%s protocol)', name)
            return
        except HandshakeException as e:
            _LOGGER.debug(
                'Failed to complete opening handshake as %s protocol: %r',
                name, e)
            if e.status:
                raise e
        except AbortedByUserException as e:
            raise
        except VersionException as e:
            raise

    # TODO(toyoshim): Add a test to cover the case all handshakers fail.
    raise HandshakeException(
        'Failed to complete opening handshake for all available protocols',
        status=common.HTTP_STATUS_BAD_REQUEST)


# vi:sts=4 sw=4 et



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from setuptools import setup

VERSION = "1.0.0"
DEPS = ["six >= 1.13.0"]

setup(
    name="mozterm",
    description="Terminal abstractions built around the blessed module.",
    license="MPL 2.0",
    author="Andrew Halberstadt",
    author_email="ahalberstadt@mozilla.com",
    url="",
    packages=["mozterm"],
    version=VERSION,
    classifiers=[
        "Environment :: Console",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Natural Language :: English",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
    ],
    install_requires=DEPS,
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import shutil
import subprocess
import sys
from pathlib import Path

import toml
from packaging.requirements import Requirement


class PoetryLockfiles:
    def __init__(
        self,
        poetry_lockfile: Path,
        pip_lockfile: Path,
    ):
        self.poetry_lockfile = poetry_lockfile
        self.pip_lockfile = pip_lockfile


class PoetryHandle:
    def __init__(self, work_dir: Path):
        self._work_dir = work_dir
        self._dependencies = {}

    def add_requirement(self, requirement: Requirement):
        self._dependencies[requirement.name] = str(requirement.specifier)

    def add_requirements_in_file(self, requirements_in: Path):
        with open(requirements_in) as requirements_in:
            for line in requirements_in.readlines():
                if line.startswith("#"):
                    continue

                req = Requirement(line)
                self.add_requirement(req)

    def reuse_existing_lockfile(self, lockfile_path: Path):
        """Make minimal number of changes to the lockfile to satisfy new requirements"""
        shutil.copy(str(lockfile_path), str(self._work_dir / "poetry.lock"))

    def generate_lockfiles(self, do_update):
        """Generate pip-style lockfiles that satisfy provided requirements

        One lockfile will be made for all mandatory requirements, and then an extra,
        compatible lockfile will be created for each optional requirement.

        Args:
            do_update: if True, then implicitly upgrade the versions of transitive
                dependencies
        """

        poetry_config = {
            "name": "poetry-test",
            "description": "",
            "version": "0",
            "authors": [],
            "dependencies": {"python": "^3.8"},
        }
        poetry_config["dependencies"].update(self._dependencies)

        pyproject = {"tool": {"poetry": poetry_config}}
        with open(self._work_dir / "pyproject.toml", "w") as pyproject_file:
            toml.dump(pyproject, pyproject_file)

        self._run_poetry(["lock"] + (["--no-update"] if not do_update else []))
        self._run_poetry(["export", "-o", "requirements.txt"])

        return PoetryLockfiles(
            self._work_dir / "poetry.lock",
            self._work_dir / "requirements.txt",
        )

    def _run_poetry(self, args):
        subprocess.check_call(
            [sys.executable, "-m", "poetry"] + args, cwd=self._work_dir
        )



from collections import namedtuple

import py
import os, sys
from py._io import terminalwriter
import codecs
import pytest

def test_get_terminal_width():
    x = py.io.get_terminal_width
    assert x == terminalwriter.get_terminal_width

def test_getdimensions(monkeypatch):
    if sys.version_info >= (3, 3):
        import shutil
        Size = namedtuple('Size', 'lines columns')
        monkeypatch.setattr(shutil, 'get_terminal_size', lambda: Size(60, 100))
        assert terminalwriter._getdimensions() == (60, 100)
    else:
        fcntl = py.test.importorskip("fcntl")
        import struct
        l = []
        monkeypatch.setattr(fcntl, 'ioctl', lambda *args: l.append(args))
        try:
            terminalwriter._getdimensions()
        except (TypeError, struct.error):
            pass
        assert len(l) == 1
        assert l[0][0] == 1

def test_terminal_width_COLUMNS(monkeypatch):
    """ Dummy test for get_terminal_width
    """
    fcntl = py.test.importorskip("fcntl")
    monkeypatch.setattr(fcntl, 'ioctl', lambda *args: int('x'))
    monkeypatch.setenv('COLUMNS', '42')
    assert terminalwriter.get_terminal_width() == 42
    monkeypatch.delenv('COLUMNS', raising=False)

def test_terminalwriter_defaultwidth_80(monkeypatch):
    monkeypatch.setattr(terminalwriter, '_getdimensions', lambda: 0/0)
    monkeypatch.delenv('COLUMNS', raising=False)
    tw = py.io.TerminalWriter()
    assert tw.fullwidth == 80

def test_terminalwriter_getdimensions_bogus(monkeypatch):
    monkeypatch.setattr(terminalwriter, '_getdimensions', lambda: (10,10))
    monkeypatch.delenv('COLUMNS', raising=False)
    tw = py.io.TerminalWriter()
    assert tw.fullwidth == 80

def test_terminalwriter_getdimensions_emacs(monkeypatch):
    # emacs terminal returns (0,0) but set COLUMNS properly
    monkeypatch.setattr(terminalwriter, '_getdimensions', lambda: (0,0))
    monkeypatch.setenv('COLUMNS', '42')
    tw = py.io.TerminalWriter()
    assert tw.fullwidth == 42

def test_terminalwriter_computes_width(monkeypatch):
    monkeypatch.setattr(terminalwriter, 'get_terminal_width', lambda: 42)
    tw = py.io.TerminalWriter()
    assert tw.fullwidth == 42

def test_terminalwriter_default_instantiation():
    tw = py.io.TerminalWriter(stringio=True)
    assert hasattr(tw, 'stringio')

def test_terminalwriter_dumb_term_no_markup(monkeypatch):
    monkeypatch.setattr(os, 'environ', {'TERM': 'dumb', 'PATH': ''})
    class MyFile:
        closed = False
        def isatty(self):
            return True
    monkeypatch.setattr(sys, 'stdout', MyFile())
    try:
        assert sys.stdout.isatty()
        tw = py.io.TerminalWriter()
        assert not tw.hasmarkup
    finally:
        monkeypatch.undo()

def test_terminalwriter_file_unicode(tmpdir):
    f = codecs.open(str(tmpdir.join("xyz")), "wb", "utf8")
    tw = py.io.TerminalWriter(file=f)
    assert tw.encoding == "utf8"

def test_unicode_encoding():
    msg = py.builtin._totext('b\u00f6y', 'utf8')
    for encoding in 'utf8', 'latin1':
        l = []
        tw = py.io.TerminalWriter(l.append, encoding=encoding)
        tw.line(msg)
        assert l[0].strip() == msg.encode(encoding)

@pytest.mark.parametrize("encoding", ["ascii"])
def test_unicode_on_file_with_ascii_encoding(tmpdir, monkeypatch, encoding):
    msg = py.builtin._totext('hell\xf6', "latin1")
    #pytest.raises(UnicodeEncodeError, lambda: bytes(msg))
    f = codecs.open(str(tmpdir.join("x")), "w", encoding)
    tw = py.io.TerminalWriter(f)
    tw.line(msg)
    f.close()
    s = tmpdir.join("x").open("rb").read().strip()
    assert encoding == "ascii"
    assert s == msg.encode("unicode-escape")


win32 = int(sys.platform == "win32")
class TestTerminalWriter:
    def pytest_generate_tests(self, metafunc):
        if "tw" in metafunc.funcargnames:
            metafunc.addcall(id="path", param="path")
            metafunc.addcall(id="stringio", param="stringio")
            metafunc.addcall(id="callable", param="callable")
    def pytest_funcarg__tw(self, request):
        if request.param == "path":
            tmpdir = request.getfuncargvalue("tmpdir")
            p = tmpdir.join("tmpfile")
            f = codecs.open(str(p), 'w+', encoding='utf8')
            tw = py.io.TerminalWriter(f)
            def getlines():
                tw._file.flush()
                return codecs.open(str(p), 'r',
                    encoding='utf8').readlines()
        elif request.param == "stringio":
            tw = py.io.TerminalWriter(stringio=True)
            def getlines():
                tw.stringio.seek(0)
                return tw.stringio.readlines()
        elif request.param == "callable":
            writes = []
            tw = py.io.TerminalWriter(writes.append)
            def getlines():
                io = py.io.TextIO()
                io.write("".join(writes))
                io.seek(0)
                return io.readlines()
        tw.getlines = getlines
        tw.getvalue = lambda: "".join(getlines())
        return tw

    def test_line(self, tw):
        tw.line("hello")
        l = tw.getlines()
        assert len(l) == 1
        assert l[0] == "hello\n"

    def test_line_unicode(self, tw):
        for encoding in 'utf8', 'latin1':
            tw._encoding = encoding
            msg = py.builtin._totext('b\u00f6y', 'utf8')
            tw.line(msg)
            l = tw.getlines()
            assert l[0] == msg + "\n"

    def test_sep_no_title(self, tw):
        tw.sep("-", fullwidth=60)
        l = tw.getlines()
        assert len(l) == 1
        assert l[0] == "-" * (60-win32) + "\n"

    def test_sep_with_title(self, tw):
        tw.sep("-", "hello", fullwidth=60)
        l = tw.getlines()
        assert len(l) == 1
        assert l[0] == "-" * 26 + " hello " + "-" * (27-win32) + "\n"

    def test_sep_longer_than_width(self, tw):
        tw.sep('-', 'a' * 10, fullwidth=5)
        line, = tw.getlines()
        # even though the string is wider than the line, still have a separator
        assert line == '- aaaaaaaaaa -\n'

    @py.test.mark.skipif("sys.platform == 'win32'")
    def test__escaped(self, tw):
        text2 = tw._escaped("hello", (31))
        assert text2.find("hello") != -1

    @py.test.mark.skipif("sys.platform == 'win32'")
    def test_markup(self, tw):
        for bold in (True, False):
            for color in ("red", "green"):
                text2 = tw.markup("hello", **{color: True, 'bold': bold})
                assert text2.find("hello") != -1
        py.test.raises(ValueError, "tw.markup('x', wronkw=3)")
        py.test.raises(ValueError, "tw.markup('x', wronkw=0)")

    def test_line_write_markup(self, tw):
        tw.hasmarkup = True
        tw.line("x", bold=True)
        tw.write("x\n", red=True)
        l = tw.getlines()
        if sys.platform != "win32":
            assert len(l[0]) >= 2, l
            assert len(l[1]) >= 2, l

    def test_attr_fullwidth(self, tw):
        tw.sep("-", "hello", fullwidth=70)
        tw.fullwidth = 70
        tw.sep("-", "hello")
        l = tw.getlines()
        assert len(l[0]) == len(l[1])

    def test_reline(self, tw):
        tw.line("hello")
        tw.hasmarkup = False
        pytest.raises(ValueError, lambda: tw.reline("x"))
        tw.hasmarkup = True
        tw.reline("0 1 2")
        tw.getlines()
        l = tw.getvalue().split("\n")
        assert len(l) == 2
        tw.reline("0 1 3")
        l = tw.getvalue().split("\n")
        assert len(l) == 2
        assert l[1].endswith("0 1 3\r")
        tw.line("so")
        l = tw.getvalue().split("\n")
        assert len(l) == 3
        assert l[-1] == ""
        assert l[1] == ("0 1 2\r0 1 3\rso   ")
        assert l[0] == "hello"


def test_terminal_with_callable_write_and_flush():
    l = set()
    class fil:
        flush = lambda self: l.add("1")
        write = lambda self, x: l.add("1")
        __call__ = lambda self, x: l.add("2")

    tw = py.io.TerminalWriter(fil())
    tw.line("hello")
    assert l == set(["1"])
    del fil.flush
    l.clear()
    tw = py.io.TerminalWriter(fil())
    tw.line("hello")
    assert l == set(["2"])


def test_chars_on_current_line():
    tw = py.io.TerminalWriter(stringio=True)

    written = []

    def write_and_check(s, expected):
        tw.write(s, bold=True)
        written.append(s)
        assert tw.chars_on_current_line == expected
        assert tw.stringio.getvalue() == ''.join(written)

    write_and_check('foo', 3)
    write_and_check('bar', 6)
    write_and_check('\n', 0)
    write_and_check('\n', 0)
    write_and_check('\n\n\n', 0)
    write_and_check('\nfoo', 3)
    write_and_check('\nfbar\nhello', 5)
    write_and_check('10', 7)


@pytest.mark.skipif(sys.platform == "win32", reason="win32 has no native ansi")
def test_attr_hasmarkup():
    tw = py.io.TerminalWriter(stringio=True)
    assert not tw.hasmarkup
    tw.hasmarkup = True
    tw.line("hello", bold=True)
    s = tw.stringio.getvalue()
    assert len(s) > len("hello\n")
    assert '\x1b[1m' in s
    assert '\x1b[0m' in s

@pytest.mark.skipif(sys.platform == "win32", reason="win32 has no native ansi")
def test_ansi_print():
    # we have no easy way to construct a file that
    # represents a terminal
    f = py.io.TextIO()
    f.isatty = lambda: True
    py.io.ansi_print("hello", 0x32, file=f)
    text2 = f.getvalue()
    assert text2.find("hello") != -1
    assert len(text2) >= len("hello\n")
    assert '\x1b[50m' in text2
    assert '\x1b[0m' in text2

def test_should_do_markup_PY_COLORS_eq_1(monkeypatch):
    monkeypatch.setitem(os.environ, 'PY_COLORS', '1')
    tw = py.io.TerminalWriter(stringio=True)
    assert tw.hasmarkup
    tw.line("hello", bold=True)
    s = tw.stringio.getvalue()
    assert len(s) > len("hello\n")
    assert '\x1b[1m' in s
    assert '\x1b[0m' in s

def test_should_do_markup_PY_COLORS_eq_0(monkeypatch):
    monkeypatch.setitem(os.environ, 'PY_COLORS', '0')
    f = py.io.TextIO()
    f.isatty = lambda: True
    tw = py.io.TerminalWriter(file=f)
    assert not tw.hasmarkup
    tw.line("hello", bold=True)
    s = f.getvalue()
    assert s == "hello\n"

def test_should_do_markup(monkeypatch):
    monkeypatch.delenv("PY_COLORS", raising=False)
    monkeypatch.delenv("NO_COLOR", raising=False)

    should_do_markup = terminalwriter.should_do_markup

    f = py.io.TextIO()
    f.isatty = lambda: True

    assert should_do_markup(f) is True

    # NO_COLOR without PY_COLORS.
    monkeypatch.setenv("NO_COLOR", "0")
    assert should_do_markup(f) is False
    monkeypatch.setenv("NO_COLOR", "1")
    assert should_do_markup(f) is False
    monkeypatch.setenv("NO_COLOR", "any")
    assert should_do_markup(f) is False

    # PY_COLORS overrides NO_COLOR ("0" and "1" only).
    monkeypatch.setenv("PY_COLORS", "1")
    assert should_do_markup(f) is True
    monkeypatch.setenv("PY_COLORS", "0")
    assert should_do_markup(f) is False
    # Uses NO_COLOR.
    monkeypatch.setenv("PY_COLORS", "any")
    assert should_do_markup(f) is False
    monkeypatch.delenv("NO_COLOR")
    assert should_do_markup(f) is True

    # Back to defaults.
    monkeypatch.delenv("PY_COLORS")
    assert should_do_markup(f) is True
    f.isatty = lambda: False
    assert should_do_markup(f) is False



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from six.moves.urllib.parse import quote

from marionette_driver.by import By
from marionette_harness import MarionetteTestCase


def inline(doc):
    return "data:text/html;charset=utf-8,{}".format(quote(doc))


class TestElementSize(MarionetteTestCase):
    def test_payload(self):
        self.marionette.navigate(inline("""<a href="#">link</a>"""))
        rect = self.marionette.find_element(By.LINK_TEXT, "link").rect
        self.assertTrue(rect["x"] > 0)
        self.assertTrue(rect["y"] > 0)
        self.assertTrue(rect["width"] > 0)
        self.assertTrue(rect["height"] > 0)



#!/usr/bin/env python

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import shutil
import sys

import mozunit
import pytest
from moztest.selftest.fixtures import binary_fixture  # noqa: F401
from mozversion import errors, get_version

"""test getting application version information from a binary path"""


@pytest.fixture()
def fake_binary(tmpdir):
    binary = tmpdir.join("binary")
    binary.write("foobar")
    return str(binary)


@pytest.fixture(name="application_ini")
def fixture_application_ini(tmpdir):
    ini = tmpdir.join("application.ini")
    ini.write(
        """[App]
ID = AppID
Name = AppName
CodeName = AppCodeName
Version = AppVersion
BuildID = AppBuildID
SourceRepository = AppSourceRepo
SourceStamp = AppSourceStamp
Vendor = AppVendor"""
    )
    return str(ini)


@pytest.fixture(name="platform_ini")
def fixture_platform_ini(tmpdir):
    ini = tmpdir.join("platform.ini")
    ini.write(
        """[Build]
BuildID = PlatformBuildID
Milestone = PlatformMilestone
SourceStamp = PlatformSourceStamp
SourceRepository = PlatformSourceRepo"""
    )
    return str(ini)


def test_real_binary(binary):  # noqa: F811
    if not binary:
        pytest.skip("No binary found")
    v = get_version(binary)
    assert isinstance(v, dict)


def test_binary(fake_binary, application_ini, platform_ini):
    _check_version(get_version(fake_binary))


@pytest.mark.skipif(
    not hasattr(os, "symlink"), reason="os.symlink not supported on this platform"
)
def test_symlinked_binary(fake_binary, application_ini, platform_ini, tmpdir):
    # create a symlink of the binary in another directory and check
    # version against this symlink
    symlink = str(tmpdir.join("symlink"))
    os.symlink(fake_binary, symlink)
    _check_version(get_version(symlink))


def test_binary_in_current_path(fake_binary, application_ini, platform_ini, tmpdir):
    os.chdir(str(tmpdir))
    _check_version(get_version())


def test_with_ini_files_on_osx(
    fake_binary, application_ini, platform_ini, monkeypatch, tmpdir
):
    monkeypatch.setattr(sys, "platform", "darwin")
    # get_version is working with ini files next to the binary
    _check_version(get_version(binary=fake_binary))

    # or if they are in the Resources dir
    # in this case the binary must be in a Contents dir, next
    # to the Resources dir
    contents_dir = tmpdir.mkdir("Contents")
    moved_binary = str(contents_dir.join(os.path.basename(fake_binary)))
    shutil.move(fake_binary, moved_binary)

    resources_dir = str(tmpdir.mkdir("Resources"))
    shutil.move(application_ini, resources_dir)
    shutil.move(platform_ini, resources_dir)

    _check_version(get_version(binary=moved_binary))


def test_invalid_binary_path(tmpdir):
    with pytest.raises(IOError):
        get_version(str(tmpdir.join("invalid")))


def test_without_ini_files(fake_binary):
    """With missing ini files an exception should be thrown"""
    with pytest.raises(errors.AppNotFoundError):
        get_version(fake_binary)


def test_without_platform_ini_file(fake_binary, application_ini):
    """With a missing platform.ini file an exception should be thrown"""
    with pytest.raises(errors.AppNotFoundError):
        get_version(fake_binary)


def test_without_application_ini_file(fake_binary, platform_ini):
    """With a missing application.ini file an exception should be thrown"""
    with pytest.raises(errors.AppNotFoundError):
        get_version(fake_binary)


def test_with_exe(application_ini, platform_ini, tmpdir):
    """Test that we can resolve .exe files"""
    binary = tmpdir.join("binary.exe")
    binary.write("foobar")
    _check_version(get_version(os.path.splitext(str(binary))[0]))


def test_not_found_with_binary_specified(fake_binary):
    with pytest.raises(errors.LocalAppNotFoundError):
        get_version(fake_binary)


def _check_version(version):
    assert version.get("application_id") == "AppID"
    assert version.get("application_name") == "AppName"
    assert version.get("application_display_name") == "AppCodeName"
    assert version.get("application_version") == "AppVersion"
    assert version.get("application_buildid") == "AppBuildID"
    assert version.get("application_repository") == "AppSourceRepo"
    assert version.get("application_changeset") == "AppSourceStamp"
    assert version.get("application_vendor") == "AppVendor"
    assert version.get("platform_name") is None
    assert version.get("platform_buildid") == "PlatformBuildID"
    assert version.get("platform_repository") == "PlatformSourceRepo"
    assert version.get("platform_changeset") == "PlatformSourceStamp"
    assert version.get("invalid_key") is None
    assert version.get("platform_version") == "PlatformMilestone"


if __name__ == "__main__":
    mozunit.main()



import os.path
import subprocess
import sys
import textwrap
from contextlib import contextmanager
from pathlib import Path
from string import ascii_lowercase

from _pytest.pytester import Pytester


@contextmanager
def subst_path_windows(filepath: Path):
    for c in ascii_lowercase[7:]:  # Create a subst drive from H-Z.
        c += ":"
        if not os.path.exists(c):
            drive = c
            break
    else:
        raise AssertionError("Unable to find suitable drive letter for subst.")

    directory = filepath.parent
    basename = filepath.name

    args = ["subst", drive, str(directory)]
    subprocess.check_call(args)
    assert os.path.exists(drive)
    try:
        filename = Path(drive, os.sep, basename)
        yield filename
    finally:
        args = ["subst", "/D", drive]
        subprocess.check_call(args)


@contextmanager
def subst_path_linux(filepath: Path):
    directory = filepath.parent
    basename = filepath.name

    target = directory / ".." / "sub2"
    os.symlink(str(directory), str(target), target_is_directory=True)
    try:
        filename = target / basename
        yield filename
    finally:
        # We don't need to unlink (it's all in the tempdir).
        pass


def test_link_resolve(pytester: Pytester) -> None:
    """See: https://github.com/pytest-dev/pytest/issues/5965."""
    sub1 = pytester.mkpydir("sub1")
    p = sub1.joinpath("test_foo.py")
    p.write_text(
        textwrap.dedent(
            """
        import pytest
        def test_foo():
            raise AssertionError()
        """
        )
    )

    subst = subst_path_linux
    if sys.platform == "win32":
        subst = subst_path_windows

    with subst(p) as subst_p:
        result = pytester.runpytest(str(subst_p), "-v")
        # i.e.: Make sure that the error is reported as a relative path, not as a
        # resolved path.
        # See: https://github.com/pytest-dev/pytest/issues/5965
        stdout = result.stdout.str()
        assert "sub1/test_foo.py" not in stdout

        # i.e.: Expect drive on windows because we just have drive:filename, whereas
        # we expect a relative path on Linux.
        expect = f"*{subst_p}*" if sys.platform == "win32" else "*sub2/test_foo.py*"
        result.stdout.fnmatch_lines([expect])



# -*- coding: utf-8 -*-
import pytest

from hpack import InvalidTableIndex
from hpack.table import HeaderTable, table_entry_size


class TestPackageFunctions:
    def test_table_entry_size(self):
        res = table_entry_size(b'TestName', b'TestValue')
        assert res == 49


class TestHeaderTable:
    def test_get_by_index_dynamic_table(self):
        tbl = HeaderTable()
        off = len(HeaderTable.STATIC_TABLE)
        val = (b'TestName', b'TestValue')
        tbl.add(*val)
        res = tbl.get_by_index(off + 1)
        assert res == val

    def test_get_by_index_static_table(self):
        tbl = HeaderTable()
        exp = (b':authority', b'')
        res = tbl.get_by_index(1)
        assert res == exp
        idx = len(HeaderTable.STATIC_TABLE)
        exp = (b'www-authenticate', b'')
        res = tbl.get_by_index(idx)
        assert res == exp

    def test_get_by_index_zero_index(self):
        tbl = HeaderTable()
        with pytest.raises(InvalidTableIndex):
            tbl.get_by_index(0)

    def test_get_by_index_out_of_range(self):
        tbl = HeaderTable()
        off = len(HeaderTable.STATIC_TABLE)
        tbl.add(b'TestName', b'TestValue')
        with pytest.raises(InvalidTableIndex) as e:
            tbl.get_by_index(off + 2)

        assert (
            "Invalid table index %d" % (off + 2) in str(e.value)
        )

    def test_repr(self):
        tbl = HeaderTable()
        tbl.add(b'TestName1', b'TestValue1')
        tbl.add(b'TestName2', b'TestValue2')
        tbl.add(b'TestName2', b'TestValue2')
        exp = (
            "HeaderTable(4096, False, deque(["
            "(b'TestName2', b'TestValue2'), "
            "(b'TestName2', b'TestValue2'), "
            "(b'TestName1', b'TestValue1')"
            "]))"
        )
        res = repr(tbl)
        assert res == exp

    def test_add_to_large(self):
        tbl = HeaderTable()
        # Max size to small to hold the value we specify
        tbl.maxsize = 1
        tbl.add(b'TestName', b'TestValue')
        # Table length should be 0
        assert len(tbl.dynamic_entries) == 0

    def test_search_in_static_full(self):
        tbl = HeaderTable()
        itm = (b':authority', b'')
        exp = (1, itm[0], itm[1])
        res = tbl.search(itm[0], itm[1])
        assert res == exp

    def test_search_in_static_partial(self):
        tbl = HeaderTable()
        exp = (1, b':authority', None)
        res = tbl.search(b':authority', b'NotInTable')
        assert res == exp

    def test_search_in_dynamic_full(self):
        tbl = HeaderTable()
        idx = len(HeaderTable.STATIC_TABLE) + 1
        tbl.add(b'TestName', b'TestValue')
        exp = (idx, b'TestName', b'TestValue')
        res = tbl.search(b'TestName', b'TestValue')
        assert res == exp

    def test_search_in_dynamic_partial(self):
        tbl = HeaderTable()
        idx = len(HeaderTable.STATIC_TABLE) + 1
        tbl.add(b'TestName', b'TestValue')
        exp = (idx, b'TestName', None)
        res = tbl.search(b'TestName', b'NotInTable')
        assert res == exp

    def test_search_no_match(self):
        tbl = HeaderTable()
        tbl.add(b'TestName', b'TestValue')
        res = tbl.search(b'NotInTable', b'NotInTable')
        assert res is None

    def test_maxsize_prop_getter(self):
        tbl = HeaderTable()
        assert tbl.maxsize == HeaderTable.DEFAULT_SIZE

    def test_maxsize_prop_setter(self):
        tbl = HeaderTable()
        exp = int(HeaderTable.DEFAULT_SIZE / 2)
        tbl.maxsize = exp
        assert tbl.resized is True
        assert tbl.maxsize == exp
        tbl.resized = False
        tbl.maxsize = exp
        assert tbl.resized is False
        assert tbl.maxsize == exp

    def test_size(self):
        tbl = HeaderTable()
        for i in range(3):
            tbl.add(b'TestName', b'TestValue')
        res = tbl._current_size
        assert res == 147

    def test_shrink_maxsize_is_zero(self):
        tbl = HeaderTable()
        tbl.add(b'TestName', b'TestValue')
        assert len(tbl.dynamic_entries) == 1
        tbl.maxsize = 0
        assert len(tbl.dynamic_entries) == 0

    def test_shrink_maxsize(self):
        tbl = HeaderTable()
        for i in range(3):
            tbl.add(b'TestName', b'TestValue')

        assert tbl._current_size == 147
        tbl.maxsize = 146
        assert len(tbl.dynamic_entries) == 2
        assert tbl._current_size == 98



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import json
import os
import signal
import socket
import sys
import time

import mozinfo
import six
from mozprocess import ProcessHandler

from mozproxy.backends.base import Playback
from mozproxy.recordings import RecordingFile
from mozproxy.utils import (
    LOG,
    download_file_from_url,
    get_available_port,
    tooltool_download,
    transform_platform,
)

here = os.path.dirname(__file__)
mitm_folder = os.path.dirname(os.path.realpath(__file__))

# maximal allowed runtime of a mitmproxy command
MITMDUMP_COMMAND_TIMEOUT = 30


class Mitmproxy(Playback):
    def __init__(self, config):
        self.config = config

        self.host = (
            "127.0.0.1" if "localhost" in self.config["host"] else self.config["host"]
        )
        self.port = None
        self.mitmproxy_proc = None
        self.mitmdump_path = None
        self.record_mode = config.get("record", False)
        self.recording = None
        self.playback_files = []

        self.browser_path = ""
        if config.get("binary", None):
            self.browser_path = os.path.normpath(config.get("binary"))

        self.policies_dir = None
        self.ignore_mitmdump_exit_failure = config.get(
            "ignore_mitmdump_exit_failure", False
        )

        if self.record_mode:
            if "recording_file" not in self.config:
                LOG.error(
                    "recording_file value was not provided. Proxy service wont' start "
                )
                raise Exception("Please provide a playback_files list.")

            if not isinstance(self.config.get("recording_file"), six.string_types):
                LOG.error("recording_file argument type is not str!")
                raise Exception("recording_file argument type invalid!")

            if not os.path.splitext(self.config.get("recording_file"))[1] == ".zip":
                LOG.error(
                    "Recording file type (%s) should be a zip. "
                    "Please provide a valid file type!"
                    % self.config.get("recording_file")
                )
                raise Exception("Recording file type should be a zip")

            if os.path.exists(self.config.get("recording_file")):
                LOG.error(
                    "Recording file (%s) already exists."
                    "Please provide a valid file path!"
                    % self.config.get("recording_file")
                )
                raise Exception("Recording file already exists.")

            if self.config.get("playback_files", False):
                LOG.error("Record mode is True and playback_files where provided!")
                raise Exception("playback_files specified during record!")

        if self.config.get("playback_version") is None:
            LOG.error(
                "mitmproxy was not provided with a 'playback_version' "
                "Please provide a valid playback version"
            )
            raise Exception("playback_version not specified!")

        # mozproxy_dir is where we will download all mitmproxy required files
        # when running locally it comes from obj_path via mozharness/mach
        if self.config.get("obj_path") is not None:
            self.mozproxy_dir = self.config.get("obj_path")
        else:
            # in production it is ../tasks/task_N/build/, in production that dir
            # is not available as an envvar, however MOZ_UPLOAD_DIR is set as
            # ../tasks/task_N/build/blobber_upload_dir so take that and go up 1 level
            self.mozproxy_dir = os.path.dirname(
                os.path.dirname(os.environ["MOZ_UPLOAD_DIR"])
            )

        self.mozproxy_dir = os.path.join(self.mozproxy_dir, "testing", "mozproxy")
        self.upload_dir = os.environ.get("MOZ_UPLOAD_DIR", self.mozproxy_dir)

        LOG.info(
            "mozproxy_dir used for mitmproxy downloads and exe files: %s"
            % self.mozproxy_dir
        )
        # setting up the MOZPROXY_DIR env variable so custom scripts know
        # where to get the data
        os.environ["MOZPROXY_DIR"] = self.mozproxy_dir

        LOG.info("Playback tool: %s" % self.config["playback_tool"])
        LOG.info("Playback tool version: %s" % self.config["playback_version"])

    def download_mitm_bin(self):
        # Download and setup mitm binaries

        manifest = os.path.join(
            here,
            "manifests",
            "mitmproxy-rel-bin-%s-{platform}.manifest"
            % self.config["playback_version"],
        )
        transformed_manifest = transform_platform(manifest, self.config["platform"])

        # generate the mitmdump_path
        self.mitmdump_path = os.path.normpath(
            os.path.join(
                self.mozproxy_dir,
                "mitmdump-%s" % self.config["playback_version"],
                "mitmdump",
            )
        )

        # Check if mitmproxy bin exists
        if os.path.exists(self.mitmdump_path):
            LOG.info("mitmproxy binary already exists. Skipping download")
        else:
            # Download and unpack mitmproxy binary
            download_path = os.path.dirname(self.mitmdump_path)
            LOG.info("create mitmproxy %s dir" % self.config["playback_version"])
            if not os.path.exists(download_path):
                os.makedirs(download_path)

            LOG.info("downloading mitmproxy binary")
            tooltool_download(
                transformed_manifest, self.config["run_local"], download_path
            )

    def download_manifest_file(self, manifest_path):
        # Manifest File
        # we use one pageset for all platforms
        LOG.info("downloading mitmproxy pageset")

        tooltool_download(manifest_path, self.config["run_local"], self.mozproxy_dir)

        with open(manifest_path) as manifest_file:
            manifest = json.load(manifest_file)
            for file in manifest:
                zip_path = os.path.join(self.mozproxy_dir, file["filename"])
                LOG.info("Adding %s to recording list" % zip_path)
                self.playback_files.append(RecordingFile(zip_path))

    def download_playback_files(self):
        # Detect type of file from playback_files and download accordingly
        if "playback_files" not in self.config:
            LOG.error(
                "playback_files value was not provided. Proxy service wont' start "
            )
            raise Exception("Please provide a playback_files list.")

        if not isinstance(self.config["playback_files"], list):
            LOG.error("playback_files should be a list")
            raise Exception("playback_files should be a list")

        for playback_file in self.config["playback_files"]:
            if playback_file.startswith("https://") and "mozilla.com" in playback_file:
                # URL provided
                dest = os.path.join(self.mozproxy_dir, os.path.basename(playback_file))
                download_file_from_url(playback_file, self.mozproxy_dir, extract=False)
                # Add Downloaded file to playback_files list
                LOG.info("Adding %s to recording list" % dest)
                self.playback_files.append(RecordingFile(dest))
                continue

            if not os.path.exists(playback_file):
                LOG.error(
                    "Zip or manifest file path (%s) does not exist. Please provide a valid path!"
                    % playback_file
                )
                raise Exception("Zip or manifest file path does not exist")

            if os.path.splitext(playback_file)[1] == ".zip":
                # zip file path provided
                LOG.info("Adding %s to recording list" % playback_file)
                self.playback_files.append(RecordingFile(playback_file))
            elif os.path.splitext(playback_file)[1] == ".manifest":
                # manifest file path provided
                self.download_manifest_file(playback_file)

    def download(self):
        """Download and unpack mitmproxy binary and pageset using tooltool"""
        if not os.path.exists(self.mozproxy_dir):
            os.makedirs(self.mozproxy_dir)

        self.download_mitm_bin()

        if self.record_mode:
            self.recording = RecordingFile(self.config["recording_file"])
        else:
            self.download_playback_files()

    def stop(self):
        LOG.info("Mitmproxy stop!!")
        self.stop_mitmproxy_playback()
        if self.record_mode:
            LOG.info("Record mode ON. Generating zip file ")
            self.recording.generate_zip_file()

    def wait(self, timeout=1):
        """Wait until the mitmproxy process has terminated."""
        # We wait using this method to allow Windows to respond to the Ctrl+Break
        # signal so that we can exit cleanly from the command-line driver.
        while True:
            returncode = self.mitmproxy_proc.wait(timeout)
            if returncode is not None:
                return returncode

    def start(self):
        # go ahead and download and setup mitmproxy
        self.download()

        # mitmproxy must be started before setup, so that the CA cert is available
        self.start_mitmproxy(self.mitmdump_path, self.browser_path)

        # In case the setup fails, we want to stop the process before raising.
        try:
            self.setup()
        except Exception:
            try:
                self.stop()
            except Exception:
                LOG.error("MitmProxy failed to STOP.", exc_info=True)
            LOG.error("Setup of MitmProxy failed.", exc_info=True)
            raise

    def start_mitmproxy(self, mitmdump_path, browser_path):
        """Startup mitmproxy and replay the specified flow file"""
        if self.mitmproxy_proc is not None:
            raise Exception("Proxy already started.")
        self.port = get_available_port()

        LOG.info("mitmdump path: %s" % mitmdump_path)
        LOG.info("browser path: %s" % browser_path)

        # mitmproxy needs some DLL's that are a part of Firefox itself, so add to path
        env = os.environ.copy()
        env["PATH"] = os.path.dirname(browser_path) + os.pathsep + env["PATH"]
        command = [mitmdump_path]

        if self.config.get("verbose", False):
            # Generate mitmproxy verbose logs
            command.extend(["-v"])
        # add proxy host and port options
        command.extend(["--listen-host", self.host, "--listen-port", str(self.port)])

        # record mode
        if self.record_mode:
            # generate recording script paths

            command.extend(
                [
                    "--save-stream-file",
                    os.path.normpath(self.recording.recording_path),
                    "--set",
                    "websocket=false",
                ]
            )
            if "inject_deterministic" in self.config.keys():
                command.extend(
                    [
                        "--scripts",
                        os.path.join(mitm_folder, "scripts", "inject-deterministic.py"),
                    ]
                )
            self.recording.set_metadata(
                "proxy_version", self.config["playback_version"]
            )
        else:
            # playback mode
            if len(self.playback_files) > 0:
                if self.config["playback_version"] == "8.1.1":
                    command.extend(
                        [
                            "--set",
                            "websocket=false",
                            "--set",
                            "connection_strategy=lazy",
                            "--set",
                            "alt_server_replay_nopop=true",
                            "--set",
                            "alt_server_replay_kill_extra=true",
                            "--set",
                            "alt_server_replay_order_reversed=true",
                            "--set",
                            "alt_server_replay={}".format(
                                ",".join(
                                    [
                                        os.path.normpath(playback_file.recording_path)
                                        for playback_file in self.playback_files
                                    ]
                                )
                            ),
                            "--scripts",
                            os.path.normpath(
                                os.path.join(
                                    mitm_folder, "scripts", "alt-serverplayback.py"
                                )
                            ),
                        ]
                    )
                elif self.config["playback_version"] in [
                    "4.0.4",
                    "5.1.1",
                    "6.0.2",
                ]:
                    command.extend(
                        [
                            "--set",
                            "upstream_cert=false",
                            "--set",
                            "upload_dir=" + os.path.normpath(self.upload_dir),
                            "--set",
                            "websocket=false",
                            "--set",
                            "server_replay_files={}".format(
                                ",".join(
                                    [
                                        os.path.normpath(playback_file.recording_path)
                                        for playback_file in self.playback_files
                                    ]
                                )
                            ),
                            "--scripts",
                            os.path.normpath(
                                os.path.join(
                                    mitm_folder, "scripts", "alternate-server-replay.py"
                                )
                            ),
                        ]
                    )
                else:
                    raise Exception("Mitmproxy version is unknown!")

            else:
                raise Exception(
                    "Mitmproxy can't start playback! Playback settings missing."
                )

        # mitmproxy needs some DLL's that are a part of Firefox itself, so add to path
        env = os.environ.copy()
        if not os.path.dirname(self.browser_path) in env["PATH"]:
            env["PATH"] = os.path.dirname(self.browser_path) + os.pathsep + env["PATH"]

        LOG.info("Starting mitmproxy playback using env path: %s" % env["PATH"])
        LOG.info("Starting mitmproxy playback using command: %s" % " ".join(command))
        # to turn off mitmproxy log output, use these params for Popen:
        # Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env)
        self.mitmproxy_proc = ProcessHandler(
            command,
            logfile=os.path.join(self.upload_dir, "mitmproxy.log"),
            env=env,
            storeOutput=False,
        )
        self.mitmproxy_proc.run()
        end_time = time.time() + MITMDUMP_COMMAND_TIMEOUT

        ready = False
        while time.time() < end_time:
            ready = self.check_proxy(host=self.host, port=self.port)
            if ready:
                LOG.info(
                    "Mitmproxy playback successfully started on %s:%d as pid %d"
                    % (self.host, self.port, self.mitmproxy_proc.pid)
                )
                return
            time.sleep(0.25)

        # cannot continue as we won't be able to playback the pages
        LOG.error("Aborting: Mitmproxy process did not startup")
        self.stop_mitmproxy_playback()
        sys.exit(1)  # XXX why do we need to do that? a raise is not enough?

    def stop_mitmproxy_playback(self):
        """Stop the mitproxy server playback"""
        if self.mitmproxy_proc is None or self.mitmproxy_proc.poll() is not None:
            return
        LOG.info(
            "Stopping mitmproxy playback, killing process %d" % self.mitmproxy_proc.pid
        )
        # On Windows, mozprocess brutally kills mitmproxy with TerminateJobObject
        # The process has no chance to gracefully shutdown.
        # Here, we send the process a break event to give it a chance to wrapup.
        # See the signal handler in the alternate-server-replay-4.0.4.py script
        if mozinfo.os == "win":
            LOG.info("Sending CTRL_BREAK_EVENT to mitmproxy")
            os.kill(self.mitmproxy_proc.pid, signal.CTRL_BREAK_EVENT)
            time.sleep(2)

        exit_code = self.mitmproxy_proc.kill()
        self.mitmproxy_proc = None

        if exit_code != 0:
            if exit_code is None:
                LOG.error("Failed to kill the mitmproxy playback process")
                return

            if mozinfo.os == "win":
                from mozprocess.winprocess import (  # noqa
                    ERROR_CONTROL_C_EXIT,
                    ERROR_CONTROL_C_EXIT_DECIMAL,
                )

                if exit_code in [ERROR_CONTROL_C_EXIT, ERROR_CONTROL_C_EXIT_DECIMAL]:
                    LOG.info(
                        "Successfully killed the mitmproxy playback process"
                        " with exit code %d" % exit_code
                    )
                    return
            log_func = LOG.error
            if self.ignore_mitmdump_exit_failure:
                log_func = LOG.info
            log_func("Mitmproxy exited with error code %d" % exit_code)
        else:
            LOG.info("Successfully killed the mitmproxy playback process")

    def check_proxy(self, host, port):
        """Check that mitmproxy process is working by doing a socket call using the proxy settings
        :param host:  Host of the proxy server
        :param port: Port of the proxy server
        :return: True if the proxy service is working
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.connect((host, port))
            s.shutdown(socket.SHUT_RDWR)
            s.close()
            return True
        except socket.error:
            return False



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from marionette_driver.by import By
from marionette_harness import MarionetteTestCase, WindowManagerMixin


class TestTextChrome(WindowManagerMixin, MarionetteTestCase):
    def setUp(self):
        super(TestTextChrome, self).setUp()
        win = self.open_chrome_window("chrome://remote/content/marionette/test.xhtml")
        self.marionette.switch_to_window(win)

        self.marionette.set_context("chrome")

    def tearDown(self):
        self.close_all_windows()

        super(TestTextChrome, self).tearDown()

    def test_get_text(self):
        elem = self.marionette.find_element(By.ID, "testBox")
        self.assertEqual(elem.text, "box")

    def test_clear_text(self):
        input = self.marionette.find_element(By.ID, "textInput3")
        self.assertEqual(
            "test",
            self.marionette.execute_script("return arguments[0].value;", [input]),
        )
        input.clear()
        self.assertEqual(
            "", self.marionette.execute_script("return arguments[0].value;", [input])
        )



# https://github.com/http2jp/http2-frame-test-case

import os
import json
import pytest
from hyperframe.frame import Frame

tc_filepaths = []
root = os.path.dirname(__file__)
path = os.walk(os.path.join(root, "http2-frame-test-case"))
for dirpath, dirnames, filenames in path:
    for filename in filenames:
        if os.path.splitext(filename)[1] != ".json":
            continue
        tc_filepaths.append(
            os.path.relpath(os.path.join(dirpath, filename), root)
        )


def check_valid_frame(tc, data):  # noqa: C901
    new_frame, length = Frame.parse_frame_header(data[:9], strict=True)
    new_frame.parse_body(memoryview(data[9:9 + length]))

    assert tc["frame"]["length"] == length
    assert tc["frame"]["stream_identifier"] == new_frame.stream_id
    assert tc["frame"]["type"] == new_frame.type

    flags = 0
    for flag, flag_bit in new_frame.defined_flags:
        if flag in new_frame.flags:
            flags |= flag_bit
    assert tc["frame"]["flags"] == flags

    p = tc["frame"]["frame_payload"]
    if "header_block_fragment" in p:
        assert p["header_block_fragment"] == new_frame.data.decode()
    if "data" in p:
        assert p["data"] == new_frame.data.decode()
    if "padding" in p:
        # the padding data itself is not retained by hyperframe after parsing
        pass
    if "padding_length" in p and p["padding_length"]:
        assert p["padding_length"] == new_frame.pad_length
    if "error_code" in p:
        assert p["error_code"] == new_frame.error_code
    if "additional_debug_data" in p:
        assert p["additional_debug_data"].encode() == new_frame.additional_data
    if "last_stream_id" in p:
        assert p["last_stream_id"] == new_frame.last_stream_id
    if "stream_dependency" in p:
        assert p["stream_dependency"] or 0 == new_frame.depends_on
    if "weight" in p and p["weight"]:
        assert p["weight"] - 1 == new_frame.stream_weight
    if "exclusive" in p:
        assert (p["exclusive"] or False) == new_frame.exclusive
    if "opaque_data" in p:
        assert p["opaque_data"].encode() == new_frame.opaque_data
    if "promised_stream_id" in p:
        assert p["promised_stream_id"] == new_frame.promised_stream_id
    if "settings" in p:
        assert dict(p["settings"]) == new_frame.settings
    if "window_size_increment" in p:
        assert p["window_size_increment"] == new_frame.window_increment


class TestExternalCollection:
    @pytest.mark.parametrize('tc_filepath', tc_filepaths)
    def test(self, tc_filepath):
        with open(os.path.join(root, tc_filepath)) as f:
            tc = json.load(f)

        data = bytes.fromhex(tc["wire"])

        if tc["error"] is None and tc["frame"]:
            check_valid_frame(tc, data)
        elif tc["error"] and tc["frame"] is None:
            with pytest.raises(Exception):
                new_frame, length = Frame.parse_frame_header(
                    data[:9],
                    strict=True
                )
                new_frame.parse_body(memoryview(data[9:9 + length]))
                assert length == new_frame.body_len
        else:
            pytest.fail("unexpected test case: {} {}".format(tc_filepath, tc))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.

import os


class Logger(object):
    """
    ManifestParser needs to ensure a singleton for mozlog as documented here:

    https://firefox-source-docs.mozilla.org/mozbase/mozlog.html#mozlog-structured-logging-for-test-output

    Logging is threadsafe, with access to handlers protected by a
    threading.Lock. However it is not process-safe. This means that
    applications using multiple processes, e.g. via the multiprocessing
    module, should arrange for all logging to happen in a single process.

    The test:
    `testing/mochitest/tests/python/test_mochitest_integration.py::test_output_testfile_in_dupe_manifests`
    creates two ManifestParser instances and runs them at the same
    tripping over the condition (above) resulting in this exception:

    [task 2023-08-02T17:16:41.636Z]   File "/builds/worker/checkouts/gecko/testing/mozbase/mozlog/mozlog/handlers/base.py", line 113, in __call__
    [task 2023-08-02T17:16:41.636Z]     self.stream.write(formatted)
    [task 2023-08-02T17:16:41.636Z] ValueError: I/O operation on closed file
    """

    logger = None
    CI = False  # True if we are running in CI

    def __init__(self):
        "Lazily will create an instance of mozlog"
        pass

    def _initialize(self):
        "Creates an instance of mozlog, if needed"
        if "TASK_ID" in os.environ:
            Logger.CI = True  # We are running in CI
        else:
            Logger.CI = False
        if Logger.logger is None:
            component = "manifestparser"
            import mozlog

            Logger.logger = mozlog.get_default_logger(component)
            if Logger.logger is None:
                Logger.logger = mozlog.unstructured.getLogger(component)

    def critical(self, *args, **kwargs):
        self._initialize()
        Logger.logger.critical(*args, **kwargs)

    def debug(self, *args, **kwargs):
        self._initialize()
        Logger.logger.debug(*args, **kwargs)

    def debug_ci(self, *args, **kwargs):
        "Log to INFO level in CI else DEBUG level"
        self._initialize()
        if Logger.CI:
            Logger.logger.info(*args, **kwargs)
        else:
            Logger.logger.debug(*args, **kwargs)

    def error(self, *args, **kwargs):
        self._initialize()
        Logger.logger.error(*args, **kwargs)

    def info(self, *args, **kwargs):
        self._initialize()
        Logger.logger.info(*args, **kwargs)

    def warning(self, *args, **kwargs):
        self._initialize()
        Logger.logger.warning(*args, **kwargs)



from six.moves.urllib.parse import quote

from marionette_driver.by import By
from marionette_driver.errors import (
    DetachedShadowRootException,
    NoSuchElementException,
    NoSuchShadowRootException,
)
from marionette_driver.marionette import WebElement, ShadowRoot

from marionette_harness import MarionetteTestCase


def inline(doc):
    return "data:text/html;charset=utf-8,{}".format(quote(doc))


page_shadow_dom = inline(
    """
        <custom-element></custom-element>
        <script>
            customElements.define('custom-element',
                class extends HTMLElement {
                    constructor() {
                            super();
                            this.attachShadow({mode: 'open'}).innerHTML = `
                                <div><a href=# id=foo>full link text</a><a href=# id=bar>another link text</a></div>
                            `;
                        }
                });
        </script>"""
)


class TestShadowDOMFindElement(MarionetteTestCase):
    def setUp(self):
        MarionetteTestCase.setUp(self)
        self.marionette.timeout.implicit = 0

    def test_find_element_from_shadow_root(self):
        self.marionette.navigate(page_shadow_dom)
        custom_element = self.marionette.find_element(By.CSS_SELECTOR, "custom-element")
        shadow_root = custom_element.shadow_root
        found = shadow_root.find_element(By.CSS_SELECTOR, "a")
        self.assertIsInstance(found, WebElement)

        el = self.marionette.execute_script(
            """
            return arguments[0].shadowRoot.querySelector('a')
        """,
            [custom_element],
        )
        self.assertEqual(found, el)

    def test_unknown_element_from_shadow_root(self):
        self.marionette.navigate(page_shadow_dom)
        shadow_root = self.marionette.find_element(
            By.CSS_SELECTOR, "custom-element"
        ).shadow_root
        with self.assertRaises(NoSuchElementException):
            shadow_root.find_element(By.CSS_SELECTOR, "does not exist")

    def test_detached_shadow_root(self):
        self.marionette.navigate(page_shadow_dom)
        shadow_root = self.marionette.find_element(
            By.CSS_SELECTOR, "custom-element"
        ).shadow_root
        self.marionette.refresh()
        with self.assertRaises(DetachedShadowRootException):
            shadow_root.find_element(By.CSS_SELECTOR, "a")

    def test_no_such_shadow_root(self):
        not_existing_shadow_root = ShadowRoot(self.marionette, "foo")
        with self.assertRaises(NoSuchShadowRootException):
            not_existing_shadow_root.find_element(By.CSS_SELECTOR, "a")


class TestShadowDOMFindElements(MarionetteTestCase):
    def setUp(self):
        MarionetteTestCase.setUp(self)
        self.marionette.timeout.implicit = 0

    def test_find_elements_from_shadow_root(self):
        self.marionette.navigate(page_shadow_dom)
        custom_element = self.marionette.find_element(By.CSS_SELECTOR, "custom-element")
        shadow_root = custom_element.shadow_root
        found = shadow_root.find_elements(By.CSS_SELECTOR, "a")
        self.assertEqual(len(found), 2)

        els = self.marionette.execute_script(
            """
            return arguments[0].shadowRoot.querySelectorAll('a')
        """,
            [custom_element],
        )

        for i in range(len(found)):
            self.assertIsInstance(found[i], WebElement)
            self.assertEqual(found[i], els[i])

    def test_detached_shadow_root(self):
        self.marionette.navigate(page_shadow_dom)
        shadow_root = self.marionette.find_element(
            By.CSS_SELECTOR, "custom-element"
        ).shadow_root
        self.marionette.refresh()
        with self.assertRaises(DetachedShadowRootException):
            shadow_root.find_elements(By.CSS_SELECTOR, "a")

    def test_no_such_shadow_root(self):
        not_existing_shadow_root = ShadowRoot(self.marionette, "foo")
        with self.assertRaises(NoSuchShadowRootException):
            not_existing_shadow_root.find_elements(By.CSS_SELECTOR, "a")



from __future__ import annotations

import dataclasses
import enum
import io
import secrets
import struct
from typing import Callable, Generator, Optional, Sequence, Tuple

from . import exceptions, extensions
from .typing import Data


try:
    from .speedups import apply_mask
except ImportError:
    from .utils import apply_mask


__all__ = [
    "Opcode",
    "OP_CONT",
    "OP_TEXT",
    "OP_BINARY",
    "OP_CLOSE",
    "OP_PING",
    "OP_PONG",
    "DATA_OPCODES",
    "CTRL_OPCODES",
    "Frame",
    "prepare_data",
    "prepare_ctrl",
    "Close",
]


class Opcode(enum.IntEnum):
    """Opcode values for WebSocket frames."""

    CONT, TEXT, BINARY = 0x00, 0x01, 0x02
    CLOSE, PING, PONG = 0x08, 0x09, 0x0A


OP_CONT = Opcode.CONT
OP_TEXT = Opcode.TEXT
OP_BINARY = Opcode.BINARY
OP_CLOSE = Opcode.CLOSE
OP_PING = Opcode.PING
OP_PONG = Opcode.PONG

DATA_OPCODES = OP_CONT, OP_TEXT, OP_BINARY
CTRL_OPCODES = OP_CLOSE, OP_PING, OP_PONG


class CloseCode(enum.IntEnum):
    """Close code values for WebSocket close frames."""

    NORMAL_CLOSURE = 1000
    GOING_AWAY = 1001
    PROTOCOL_ERROR = 1002
    UNSUPPORTED_DATA = 1003
    # 1004 is reserved
    NO_STATUS_RCVD = 1005
    ABNORMAL_CLOSURE = 1006
    INVALID_DATA = 1007
    POLICY_VIOLATION = 1008
    MESSAGE_TOO_BIG = 1009
    MANDATORY_EXTENSION = 1010
    INTERNAL_ERROR = 1011
    SERVICE_RESTART = 1012
    TRY_AGAIN_LATER = 1013
    BAD_GATEWAY = 1014
    TLS_HANDSHAKE = 1015


# See https://www.iana.org/assignments/websocket/websocket.xhtml
CLOSE_CODE_EXPLANATIONS: dict[int, str] = {
    CloseCode.NORMAL_CLOSURE: "OK",
    CloseCode.GOING_AWAY: "going away",
    CloseCode.PROTOCOL_ERROR: "protocol error",
    CloseCode.UNSUPPORTED_DATA: "unsupported data",
    CloseCode.NO_STATUS_RCVD: "no status received [internal]",
    CloseCode.ABNORMAL_CLOSURE: "abnormal closure [internal]",
    CloseCode.INVALID_DATA: "invalid frame payload data",
    CloseCode.POLICY_VIOLATION: "policy violation",
    CloseCode.MESSAGE_TOO_BIG: "message too big",
    CloseCode.MANDATORY_EXTENSION: "mandatory extension",
    CloseCode.INTERNAL_ERROR: "internal error",
    CloseCode.SERVICE_RESTART: "service restart",
    CloseCode.TRY_AGAIN_LATER: "try again later",
    CloseCode.BAD_GATEWAY: "bad gateway",
    CloseCode.TLS_HANDSHAKE: "TLS handshake failure [internal]",
}


# Close code that are allowed in a close frame.
# Using a set optimizes `code in EXTERNAL_CLOSE_CODES`.
EXTERNAL_CLOSE_CODES = {
    CloseCode.NORMAL_CLOSURE,
    CloseCode.GOING_AWAY,
    CloseCode.PROTOCOL_ERROR,
    CloseCode.UNSUPPORTED_DATA,
    CloseCode.INVALID_DATA,
    CloseCode.POLICY_VIOLATION,
    CloseCode.MESSAGE_TOO_BIG,
    CloseCode.MANDATORY_EXTENSION,
    CloseCode.INTERNAL_ERROR,
    CloseCode.SERVICE_RESTART,
    CloseCode.TRY_AGAIN_LATER,
    CloseCode.BAD_GATEWAY,
}


OK_CLOSE_CODES = {
    CloseCode.NORMAL_CLOSURE,
    CloseCode.GOING_AWAY,
    CloseCode.NO_STATUS_RCVD,
}


BytesLike = bytes, bytearray, memoryview


@dataclasses.dataclass
class Frame:
    """
    WebSocket frame.

    Attributes:
        opcode: Opcode.
        data: Payload data.
        fin: FIN bit.
        rsv1: RSV1 bit.
        rsv2: RSV2 bit.
        rsv3: RSV3 bit.

    Only these fields are needed. The MASK bit, payload length and masking-key
    are handled on the fly when parsing and serializing frames.

    """

    opcode: Opcode
    data: bytes
    fin: bool = True
    rsv1: bool = False
    rsv2: bool = False
    rsv3: bool = False

    def __str__(self) -> str:
        """
        Return a human-readable representation of a frame.

        """
        coding = None
        length = f"{len(self.data)} byte{'' if len(self.data) == 1 else 's'}"
        non_final = "" if self.fin else "continued"

        if self.opcode is OP_TEXT:
            # Decoding only the beginning and the end is needlessly hard.
            # Decode the entire payload then elide later if necessary.
            data = repr(self.data.decode())
        elif self.opcode is OP_BINARY:
            # We'll show at most the first 16 bytes and the last 8 bytes.
            # Encode just what we need, plus two dummy bytes to elide later.
            binary = self.data
            if len(binary) > 25:
                binary = b"".join([binary[:16], b"\x00\x00", binary[-8:]])
            data = " ".join(f"{byte:02x}" for byte in binary)
        elif self.opcode is OP_CLOSE:
            data = str(Close.parse(self.data))
        elif self.data:
            # We don't know if a Continuation frame contains text or binary.
            # Ping and Pong frames could contain UTF-8.
            # Attempt to decode as UTF-8 and display it as text; fallback to
            # binary. If self.data is a memoryview, it has no decode() method,
            # which raises AttributeError.
            try:
                data = repr(self.data.decode())
                coding = "text"
            except (UnicodeDecodeError, AttributeError):
                binary = self.data
                if len(binary) > 25:
                    binary = b"".join([binary[:16], b"\x00\x00", binary[-8:]])
                data = " ".join(f"{byte:02x}" for byte in binary)
                coding = "binary"
        else:
            data = "''"

        if len(data) > 75:
            data = data[:48] + "..." + data[-24:]

        metadata = ", ".join(filter(None, [coding, length, non_final]))

        return f"{self.opcode.name} {data} [{metadata}]"

    @classmethod
    def parse(
        cls,
        read_exact: Callable[[int], Generator[None, None, bytes]],
        *,
        mask: bool,
        max_size: Optional[int] = None,
        extensions: Optional[Sequence[extensions.Extension]] = None,
    ) -> Generator[None, None, Frame]:
        """
        Parse a WebSocket frame.

        This is a generator-based coroutine.

        Args:
            read_exact: generator-based coroutine that reads the requested
                bytes or raises an exception if there isn't enough data.
            mask: whether the frame should be masked i.e. whether the read
                happens on the server side.
            max_size: maximum payload size in bytes.
            extensions: list of extensions, applied in reverse order.

        Raises:
            EOFError: if the connection is closed without a full WebSocket frame.
            UnicodeDecodeError: if the frame contains invalid UTF-8.
            PayloadTooBig: if the frame's payload size exceeds ``max_size``.
            ProtocolError: if the frame contains incorrect values.

        """
        # Read the header.
        data = yield from read_exact(2)
        head1, head2 = struct.unpack("!BB", data)

        # While not Pythonic, this is marginally faster than calling bool().
        fin = True if head1 & 0b10000000 else False
        rsv1 = True if head1 & 0b01000000 else False
        rsv2 = True if head1 & 0b00100000 else False
        rsv3 = True if head1 & 0b00010000 else False

        try:
            opcode = Opcode(head1 & 0b00001111)
        except ValueError as exc:
            raise exceptions.ProtocolError("invalid opcode") from exc

        if (True if head2 & 0b10000000 else False) != mask:
            raise exceptions.ProtocolError("incorrect masking")

        length = head2 & 0b01111111
        if length == 126:
            data = yield from read_exact(2)
            (length,) = struct.unpack("!H", data)
        elif length == 127:
            data = yield from read_exact(8)
            (length,) = struct.unpack("!Q", data)
        if max_size is not None and length > max_size:
            raise exceptions.PayloadTooBig(
                f"over size limit ({length} > {max_size} bytes)"
            )
        if mask:
            mask_bytes = yield from read_exact(4)

        # Read the data.
        data = yield from read_exact(length)
        if mask:
            data = apply_mask(data, mask_bytes)

        frame = cls(opcode, data, fin, rsv1, rsv2, rsv3)

        if extensions is None:
            extensions = []
        for extension in reversed(extensions):
            frame = extension.decode(frame, max_size=max_size)

        frame.check()

        return frame

    def serialize(
        self,
        *,
        mask: bool,
        extensions: Optional[Sequence[extensions.Extension]] = None,
    ) -> bytes:
        """
        Serialize a WebSocket frame.

        Args:
            mask: whether the frame should be masked i.e. whether the write
                happens on the client side.
            extensions: list of extensions, applied in order.

        Raises:
            ProtocolError: if the frame contains incorrect values.

        """
        self.check()

        if extensions is None:
            extensions = []
        for extension in extensions:
            self = extension.encode(self)

        output = io.BytesIO()

        # Prepare the header.
        head1 = (
            (0b10000000 if self.fin else 0)
            | (0b01000000 if self.rsv1 else 0)
            | (0b00100000 if self.rsv2 else 0)
            | (0b00010000 if self.rsv3 else 0)
            | self.opcode
        )

        head2 = 0b10000000 if mask else 0

        length = len(self.data)
        if length < 126:
            output.write(struct.pack("!BB", head1, head2 | length))
        elif length < 65536:
            output.write(struct.pack("!BBH", head1, head2 | 126, length))
        else:
            output.write(struct.pack("!BBQ", head1, head2 | 127, length))

        if mask:
            mask_bytes = secrets.token_bytes(4)
            output.write(mask_bytes)

        # Prepare the data.
        if mask:
            data = apply_mask(self.data, mask_bytes)
        else:
            data = self.data
        output.write(data)

        return output.getvalue()

    def check(self) -> None:
        """
        Check that reserved bits and opcode have acceptable values.

        Raises:
            ProtocolError: if a reserved bit or the opcode is invalid.

        """
        if self.rsv1 or self.rsv2 or self.rsv3:
            raise exceptions.ProtocolError("reserved bits must be 0")

        if self.opcode in CTRL_OPCODES:
            if len(self.data) > 125:
                raise exceptions.ProtocolError("control frame too long")
            if not self.fin:
                raise exceptions.ProtocolError("fragmented control frame")


def prepare_data(data: Data) -> Tuple[int, bytes]:
    """
    Convert a string or byte-like object to an opcode and a bytes-like object.

    This function is designed for data frames.

    If ``data`` is a :class:`str`, return ``OP_TEXT`` and a :class:`bytes`
    object encoding ``data`` in UTF-8.

    If ``data`` is a bytes-like object, return ``OP_BINARY`` and a bytes-like
    object.

    Raises:
        TypeError: if ``data`` doesn't have a supported type.

    """
    if isinstance(data, str):
        return OP_TEXT, data.encode("utf-8")
    elif isinstance(data, BytesLike):
        return OP_BINARY, data
    else:
        raise TypeError("data must be str or bytes-like")


def prepare_ctrl(data: Data) -> bytes:
    """
    Convert a string or byte-like object to bytes.

    This function is designed for ping and pong frames.

    If ``data`` is a :class:`str`, return a :class:`bytes` object encoding
    ``data`` in UTF-8.

    If ``data`` is a bytes-like object, return a :class:`bytes` object.

    Raises:
        TypeError: if ``data`` doesn't have a supported type.

    """
    if isinstance(data, str):
        return data.encode("utf-8")
    elif isinstance(data, BytesLike):
        return bytes(data)
    else:
        raise TypeError("data must be str or bytes-like")


@dataclasses.dataclass
class Close:
    """
    Code and reason for WebSocket close frames.

    Attributes:
        code: Close code.
        reason: Close reason.

    """

    code: int
    reason: str

    def __str__(self) -> str:
        """
        Return a human-readable representation of a close code and reason.

        """
        if 3000 <= self.code < 4000:
            explanation = "registered"
        elif 4000 <= self.code < 5000:
            explanation = "private use"
        else:
            explanation = CLOSE_CODE_EXPLANATIONS.get(self.code, "unknown")
        result = f"{self.code} ({explanation})"

        if self.reason:
            result = f"{result} {self.reason}"

        return result

    @classmethod
    def parse(cls, data: bytes) -> Close:
        """
        Parse the payload of a close frame.

        Args:
            data: payload of the close frame.

        Raises:
            ProtocolError: if data is ill-formed.
            UnicodeDecodeError: if the reason isn't valid UTF-8.

        """
        if len(data) >= 2:
            (code,) = struct.unpack("!H", data[:2])
            reason = data[2:].decode("utf-8")
            close = cls(code, reason)
            close.check()
            return close
        elif len(data) == 0:
            return cls(CloseCode.NO_STATUS_RCVD, "")
        else:
            raise exceptions.ProtocolError("close frame too short")

    def serialize(self) -> bytes:
        """
        Serialize the payload of a close frame.

        """
        self.check()
        return struct.pack("!H", self.code) + self.reason.encode("utf-8")

    def check(self) -> None:
        """
        Check that the close code has a valid value for a close frame.

        Raises:
            ProtocolError: if the close code is invalid.

        """
        if not (self.code in EXTERNAL_CLOSE_CODES or 3000 <= self.code < 5000):
            raise exceptions.ProtocolError("invalid status code")
